import pytest

from asyncjobs.logmux import LogMux

pytestmark = pytest.mark.asyncio


async def test_no_output(verify_output):
    async with LogMux():
        pass
    assert verify_output([[]])


async def test_one_stream_undecorated(verify_output):
    async with LogMux() as logmux:
        async with logmux.new_stream() as f:
            print('This is the first line', file=f)
            print('This is the second line', file=f)
    assert verify_output(
        [['This is the first line', 'This is the second line']]
    )


async def test_two_streams_undecorated(verify_output):
    async with LogMux() as logmux:
        async with logmux.new_stream() as f:
            print('This is stream 1 line 1', file=f)
            async with logmux.new_stream() as g:
                print('This is stream 2 line 1', file=g)
                print('This is stream 2 line 2', file=g)
            print('This is stream 1 line 2', file=f)
    assert verify_output(
        [
            ['This is stream 1 line 1', 'This is stream 1 line 2'],
            ['This is stream 2 line 1', 'This is stream 2 line 2'],
        ],
    )


async def test_one_stream_decorated(verify_output):
    async with LogMux() as logmux:
        async with logmux.new_stream(lambda s: f'[pre]{s[:-1]}[post]\n') as f:
            print('This is the first line', file=f)
            print('This is the second line', file=f)
    assert verify_output(
        [
            [
                '[pre]This is the first line[post]',
                '[pre]This is the second line[post]',
            ]
        ]
    )


async def test_two_streams_decorated(verify_output):
    async with LogMux() as logmux:
        async with logmux.new_stream(lambda s: f'1>>{s[:-1]}<<1\n') as f:
            print('This is stream 1 line 1', file=f)
            async with logmux.new_stream(lambda s: f'2>>{s[:-1]}<<2\n') as g:
                print('This is stream 2 line 1', file=g)
                print('This is stream 2 line 2', file=g)
            print('This is stream 1 line 2', file=f)
    assert verify_output(
        [
            ['1>>This is stream 1 line 1<<1', '1>>This is stream 1 line 2<<1'],
            ['2>>This is stream 2 line 1<<2', '2>>This is stream 2 line 2<<2'],
        ],
    )


async def test_one_charwise_stream_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    async with LogMux() as logmux:
        async with logmux.new_stream(lambda s: f'<{s.rstrip()}>\n') as f:
            for c in s:
                f.write(c)
    assert verify_output([['<foo>', '<bar>', '<baz>']])


async def test_two_charwise_streams_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    t = '123\n456\n789'
    async with LogMux() as logmux:
        async with logmux.new_stream(lambda s: f'<{s.rstrip()}>\n') as f:
            async with logmux.new_stream(lambda s: f'[{s.rstrip()}]\n') as g:
                for c, d in zip(s, t):
                    f.write(c)
                    g.write(d)
    assert verify_output(
        [['<foo>', '<bar>', '<baz>'], ['[123]', '[456]', '[789]']]
    )


async def test_one_bytewise_stream_with_garbage(tmp_path):
    # LogMux is text/line-based, but uses surrogateescape to preserve
    # undecodable bytes from input to output. We'd like to use capfdbinary
    # here, but _pytest.capture.EncodedFile does not support surrogateescape
    # handling on output. Use a temp file instead.
    output = tmp_path / 'output'
    lines = [
        b'first line...',
        b'latin-1: \xc6\xd8\xc5...',
        b'utf-8:   \xe2\x9c\x94\xe2\x88\x80\xe2\x9c\x98...',
        b'f8 - ff: \xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff...',
        b'last line without newline',
    ]
    bytestring = b'\n'.join(lines)
    with output.open('w', errors='surrogateescape') as out:
        async with LogMux(out) as logmux:
            async with logmux.new_stream(lambda s: f'❰{s.rstrip()}❱\n') as f:
                f.buffer.write(bytestring)
    actual = output.read_text(errors='surrogateescape')
    expect_lines = [
        '❰first line...❱',
        '❰latin-1: \udcc6\udcd8\udcc5...❱',
        '❰utf-8:   ✔∀✘...❱',
        '❰f8 - ff: \udcf8\udcf9\udcfa\udcfb\udcfc\udcfd\udcfe\udcff...❱',
        '❰last line without newline❱',
    ]
    assert actual == ''.join(line + '\n' for line in expect_lines)



import contextlib
import logging
import sys

from . import external_work, logmux

logger = logging.getLogger(__name__)


class Job(external_work.Job):
    """Job with stdout/stderr redirected via a LogMux-enabled Scheduler.

    This enables jobs to have their output multiplexed to a single (pair of)
    output stream(s) controlled by the Scheduler.

    Redirection of the actual stdout/stderr file descriptors is automatically
    done for subprocesses, and for self.logger (unless redirect_logger is set
    to False). Other output (from thread workers or directly from .__call__()
    must be redirected to self.stdout/self.stderr manually.)
    """

    def __init__(self, *args, redirect_logger=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.stdout = None
        self.stderr = None
        self.redirect_logger = redirect_logger

    def decorate_out(self, msg):
        """Manipulate each message sent to self.stdout."""
        return msg

    def decorate_err(self, msg):
        """Manipulate each message sent to self.stderr."""
        return msg

    @contextlib.asynccontextmanager
    async def _setup_redirection(self, scheduler):
        async with scheduler.outmux.new_stream(self.decorate_out) as outf:
            async with scheduler.errmux.new_stream(self.decorate_err) as errf:
                self.stdout = outf
                self.stderr = errf
                log_handler = None
                if self.redirect_logger:
                    log_handler = logging.StreamHandler(self.stderr)
                    self.logger.addHandler(log_handler)
                try:
                    yield
                finally:
                    if log_handler is not None:
                        self.logger.removeHandler(log_handler)
                    self.stdout = None
                    self.stderr = None

    async def __call__(self, scheduler):
        async with self._setup_redirection(scheduler):
            return await super().__call__(scheduler)

    async def run_in_subprocess(self, argv, **kwargs):
        """Pass redirected out/err as stdout/stderr to the subprocess.

        Only if stdout/stderr is not already customized by the caller.
        """
        if kwargs.get('stdout') is None:
            kwargs['stdout'] = self.stdout
        if kwargs.get('stderr') is None:
            kwargs['stderr'] = self.stderr
        return await super().run_in_subprocess(argv, **kwargs)


class Scheduler(external_work.Scheduler):
    """Run jobs with their output multiplexed through LogMux.

    Extend Scheduler with two LogMux instances - one for stdout and one for
    stderr - which are used by the above Job class to setup their own stdout
    and stderr streams redirected through these instances.
    """

    def __init__(self, *, outmux=None, errmux=None, **kwargs):
        self.outmux = logmux.LogMux(sys.stdout) if outmux is None else outmux
        self.errmux = logmux.LogMux(sys.stderr) if errmux is None else errmux
        super().__init__(**kwargs)

    async def _run_tasks(self, *args, **kwargs):
        logger.debug('Starting LogMux instances…')
        async with self.outmux:
            async with self.errmux:
                await super()._run_tasks(*args, **kwargs)



import asyncio
from contextlib import contextmanager
from functools import partial
import logging
import os
import pytest
import signal
from subprocess import CalledProcessError
import time

from asyncjobs import basic, external_work

logger = logging.getLogger(__name__)


@contextmanager
def abort_in(when=None, assert_on_escape=True):
    """Simulate Ctrl+C (aka. KeyboardInterrupt/SIGINT) after the given delay.

    This will schedule a SIGINT signal in the current process after the given
    number of seconds if we're still within the context. Upon context exit,
    the signal is cancelled.

    If the signal is handled within the context, no further action is taken.
    Otherwise, if the signal/KeyboardInterrupt escapes the context, an error
    is logged and (unless assert_on_escape is explicitly disabled) an
    assertion is raised (to fail the current test).
    """
    if when is None:  # Do nothing
        yield
        return

    async def wait_and_kill():
        await asyncio.sleep(when)
        logger.warning('Raising SIGINT to simulate Ctrl+C…')
        os.kill(os.getpid(), signal.SIGINT)

    task = asyncio.create_task(wait_and_kill())
    try:
        yield
    except KeyboardInterrupt:
        logger.error('SIGINT/KeyboardInterrupt escaped the context!')
        if assert_on_escape:
            assert False, 'SIGINT/KeyboardInterrupt escaped the context!'
    finally:
        if not task.done():
            logger.debug(f'context complete before {when}')
            task.cancel()


@contextmanager
def assert_elapsed_time_within(time_limit):
    """Measure time used in context and fail test if not within given limit."""
    before = time.time()
    try:
        yield
    finally:
        after = time.time()
        assert after < before + time_limit


# Wildcard used for don't-care values in expected event fields
Whatever = object()


def verify_event(expect, actual):
    expect_keys = set(expect.keys()) - {'MAY_CANCEL'}
    assert expect_keys == set(actual.keys())
    keys = sorted(expect_keys)
    for e, a in zip((expect[k] for k in keys), (actual[k] for k in keys)):
        if e is not Whatever:
            assert e == a
    return True


class TSimpleJob(basic.Job):
    """Async jobs with test instrumentation."""

    def __init__(
        self,
        name,
        deps=None,
        *,
        before=None,
        call=None,
        async_sleep=0,
        spawn=None,
        await_spawn=False,
        result=None,
    ):
        super().__init__(name=name, deps=deps)
        self.before = set() if before is None else set(before)
        self.call = call
        self.async_sleep = async_sleep
        self.spawn = [] if spawn is None else spawn
        self.await_spawn = await_spawn
        self.result = '{} done'.format(name) if result is None else result

        self._expected_events = []
        self._expect_event('add'),
        self._expect_event('start'),

    def _expect_event(self, event, *, may_cancel=False, **kwargs):
        d = {'event': event, 'job': self.name, 'timestamp': Whatever}
        d.update(kwargs)
        if may_cancel:
            d['MAY_CANCEL'] = True
        self._expected_events.append(d)

    async def do_work(self, scheduler):
        result = None
        if self.call:
            result = self.call()
        return self.result if result is None else result

    def expected_events(self):
        if self._expected_events[-1]['event'] != 'finish':  # we were cancelled
            self._expect_event('finish', fate='cancelled')
        return self._expected_events

    async def __call__(self, scheduler):
        if self.deps:
            if self.deps == {'MISSING'}:
                self._expect_event('finish', fate='failed')  # expect KeyError
            else:
                self._expect_event(
                    'await results', jobs=list(self.deps), pending=Whatever
                )
        dep_results = await super().__call__(scheduler)
        if self.deps:
            self._expect_event('awaited results')
        self.logger.debug(f'Results from deps: {dep_results}')

        if self.async_sleep:
            self.logger.info(f'Async sleep for {self.async_sleep} seconds…')
            await asyncio.sleep(self.async_sleep)
            self.logger.info('Finished async sleep')

        result = await self.do_work(scheduler)

        for job in self.spawn:
            scheduler.add(job)

        if self.await_spawn and self.spawn:
            spawn = [job.name for job in self.spawn]
            self._expect_event('await results', jobs=spawn, pending=spawn)
            await scheduler.results(*[job.name for job in self.spawn])
            self._expect_event('awaited results')

        for b in self.before:
            assert b in scheduler.tasks  # The other job has been started
            assert not scheduler.tasks[b].done()  # but is not yet finished

        if isinstance(result, Exception):
            self.logger.info(f'Raising exception: {result}')
            self._expect_event('finish', fate='failed')
            raise result
        else:
            self.logger.info(f'Returning result: {result!r}')
            self._expect_event('finish', fate='success')
            return result


class TExternalWorkJob(TSimpleJob, external_work.Job):
    """Test jobs with thread/subprocess capabilities."""

    def __init__(
        self,
        *args,
        thread_sleep=0,
        thread=None,
        subproc_sleep=0,
        subproc=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if thread_sleep:
            if thread is not None:
                raise ValueError('Cannot both sleep and work in thread')
            self.thread = partial(time.sleep, thread_sleep)
        else:
            self.thread = thread
        if subproc_sleep:
            if subproc is not None:
                raise ValueError('Cannot both sleep and work in subprocess')
            self.subproc = ['sleep', str(subproc_sleep)]
        else:
            self.subproc = subproc

    async def _do_thread_stuff(self, scheduler):
        self.logger.debug(f'Await call {self.thread} in thread…')
        self._expect_event('await worker slot')
        self._expect_event('awaited worker slot', may_cancel=True)
        self._expect_event(
            'await worker thread', may_cancel=True, func=Whatever
        )
        try:
            ret = await self.call_in_thread(self.thread)
            self._expect_event('awaited worker thread', fate='success')
        except asyncio.CancelledError:
            self._expect_event(
                'awaited worker thread', may_cancel=True, fate='cancelled',
            )
            raise
        except Exception as e:
            ret = e
            self._expect_event('awaited worker thread', fate='failed')
        self.logger.debug(f'Finished thread call: {ret!r}')
        return ret

    async def _do_subproc_stuff(self, scheduler):
        self.logger.debug(f'Await run {self.subproc} in subprocess…')
        self._expect_event('await worker slot')
        self._expect_event('awaited worker slot', may_cancel=True)
        self._expect_event(
            'await worker proc', may_cancel=True, argv=self.subproc
        )
        try:
            ret = await self.run_in_subprocess(self.subproc)
            self._expect_event('awaited worker proc', exit=0)
        except asyncio.CancelledError:
            self._expect_event(
                'awaited worker proc', may_cancel=True, exit=-15
            )
            raise
        except Exception as e:
            ret = e
            if isinstance(e, CalledProcessError):
                self._expect_event('awaited worker proc', exit=e.returncode)
        self.logger.debug(f'Finished subprocess run: {ret}')
        return ret

    async def do_work(self, scheduler):
        result = await super().do_work(scheduler)
        if self.thread:
            result = await self._do_thread_stuff(scheduler)
        if self.subproc:
            result = await self._do_subproc_stuff(scheduler)
        return result


def verify_events(scheduler, start_time, actual, todo):
    after = time.time()
    num_jobs = len(todo)
    num_tasks = len(scheduler.tasks)
    expect = {j.name: j.expected_events() for j in todo}

    # Timestamps are in sorted order and all between 'start_time' and 'after'
    timestamps = [e['timestamp'] for e in actual]
    assert timestamps == sorted(timestamps)
    assert timestamps[0] >= start_time
    assert timestamps[-1] <= after

    # Initial jobs are added before execution starts
    expect_adds = [e.pop(0) for e in expect.values()]
    for e, a in zip(expect_adds, actual[:num_jobs]):
        assert verify_event(e, a)
    actual = actual[num_jobs:]

    # Overall execution start and finish
    overall_start, overall_finish = actual.pop(0), actual.pop()
    assert verify_event(
        {
            'event': 'start',
            'num_jobs': num_jobs,
            'keep_going': Whatever,
            'timestamp': Whatever,
        },
        overall_start,
    )
    assert verify_event(
        {'event': 'finish', 'num_tasks': num_tasks, 'timestamp': Whatever},
        overall_finish,
    )

    if expect:
        # Jobs are started
        expect_starts = [e.pop(0) for e in expect.values()]
        for e, a in zip(expect_starts, actual[:num_jobs]):
            assert verify_event(e, a)
        actual = actual[num_jobs:]

        # Await task execution
        await_tasks, awaited_tasks = actual.pop(0), actual.pop()
        assert verify_event(
            {
                'event': 'await tasks',
                'jobs': [j.name for j in todo],
                'timestamp': Whatever,
            },
            await_tasks,
        )
        assert verify_event(
            {'event': 'awaited tasks', 'timestamp': Whatever}, awaited_tasks,
        )

        # Remaining events belong to individual tasks
        # This includes expected events from tasks that were spawned from
        # other tasks (hence not part of the original 'todo').
        for name, job in scheduler.jobs.items():
            if name not in expect:  # job was spawned after .run()
                expect[name] = job.expected_events()
        # Verify each event by pairing it w/that job's next expected event
        while actual:
            a = actual.pop(0)
            job_name = a['job']
            e = expect[job_name].pop(0)
            if a['event'] == 'finish' and a['fate'] == 'cancelled':
                while e.get('MAY_CANCEL', False):
                    e = expect[job_name].pop(0)
            assert verify_event(e, a)

        # No more expected events
        assert all(len(e) == 0 for e in expect.values())

    assert not actual  # no more actual events
    return True


@pytest.fixture(params=[1, 2, 4, 100])
def num_workers(request):
    return request.param


@pytest.fixture
def scheduler_with_workers(num_workers):
    def make_scheduler_class(*bases):
        class _Scheduler(*bases):
            pass

        logger.info(f'creating scheduler with {num_workers} worker threads')
        return partial(_Scheduler, workers=num_workers)

    return make_scheduler_class


@contextmanager
def setup_scheduler(scheduler_class, todo):
    events = []
    scheduler = scheduler_class(event_handler=events.append)
    before = time.time()
    for job in todo:
        scheduler.add(job)
    yield scheduler
    verify_events(scheduler, before, events, todo)


Cancelled = object()


def verify_tasks(tasks, expects):
    errors = 0

    def fail(job_name, expect, actual=None):
        nonlocal errors
        if actual is None:
            logger.error(f'{job_name}: {expect}')
        else:
            logger.error(f'{job_name}: expected {expect!r}, actual {actual!r}')
        errors += 1

    for name in set(tasks.keys()) | set(expects.keys()):
        try:
            expect = expects[name]
            task = tasks[name]
        except KeyError:
            e = 'present' if name in expects else 'missing'
            t = 'present' if name in tasks else 'missing'
            fail(name, f'{e} in expects, {t} in tasks')
            continue
        if expect is Cancelled:
            if not task.cancelled():
                fail(name, Cancelled, task)
        elif isinstance(expect, Exception):
            e = task.exception()
            if not isinstance(e, expect.__class__):
                fail(name, expect.__class__, type(e))
            if e.args != expect.args:
                fail(name, expect.args, e.args)
        else:
            if task.result() != expect:
                fail(name, expect, task.result())
    return errors == 0


@pytest.fixture
def verify_output(capfd):
    def _verify_one_output(expect_lines_from_streams, actual_text):
        logger.debug(
            f'Verifying {actual_text!r} against {expect_lines_from_streams}'
        )
        actual_lines = actual_text.split('\n')
        assert actual_lines.pop() == ''
        for actual_line in actual_lines:
            # Expect this line is the next line from one of our streams
            expected = set()
            for expect_lines in expect_lines_from_streams:
                if expect_lines and actual_line == expect_lines[0]:  # found it
                    expect_lines.pop(0)
                    break
                elif expect_lines:
                    expected.add(expect_lines[0])
            else:  # no match for any stream
                assert False, f'actual: {actual_line!r}, expect: {expected!r}'
        # no more lines expected
        assert all(e == [] for e in expect_lines_from_streams)

    def _verify_output(expect_stdout_streams, expect_stderr_streams=None):
        actual = capfd.readouterr()
        _verify_one_output(expect_stdout_streams, actual.out)
        if expect_stderr_streams is not None:
            _verify_one_output(expect_stderr_streams, actual.err)
        return True

    return _verify_output



import asyncio
import contextlib
import logging
import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory

logger = logging.getLogger(__name__)


class LogMux:
    """Async task to multiplex many write streams into a single stream."""

    def __init__(self, out=None, tmp_base=None):
        self.out = sys.stdout if out is None else out
        self.q = asyncio.Queue()
        self.tempdir = TemporaryDirectory(dir=tmp_base, prefix='LogMux_')
        self.fifonum = 0
        self._task = None

    async def _watch(self, path, decorator=None):
        """Add the given 'path' to be watched by LogMux.

        Lines read from 'path' will be passed through 'decorator' before being
        written to this logmux's shared output.
        """
        await self.q.put(('watch', path, decorator))
        await self.q.join()

    async def watched_fifo(self, decorator=None):
        """Create a FIFO (aka. named pipe) that is watched by LogMux.

        Creates a FIFO in self.tempdir, and returns its path.
        The FIFO is watched by LogMux, so anything written into it will appear
        on LogMux's output (after being passed through 'decorator').

        The FIFO (and the rest of self.tempdir) will be automatically removed
        on shutdown.
        """
        self.fifonum += 1
        fifopath = Path(self.tempdir.name, f'fifo{self.fifonum}')
        assert not fifopath.exists()
        os.mkfifo(fifopath)
        await self._watch(fifopath, decorator)
        return fifopath

    async def unwatch(self, path):
        """Stop watching the given 'path'."""
        await self.q.put(('unwatch', path))
        await self.q.join()

    @contextlib.asynccontextmanager
    async def new_stream(self, decorator=None):
        """Context manager wrapping .watched_fifo() and .unwatch()."""
        path = await self.watched_fifo(decorator)
        try:
            with open(path, 'w') as f:
                yield f
        finally:
            await self.unwatch(path)

    async def shutdown(self):
        """Shutdown LogMux. Stop watching all files and cleanup temporaries."""
        await self.q.put(('shutdown',))  # Signal shutdown
        await self.q.join()
        self.tempdir.cleanup()

    async def service(self):
        """Coroutine reading from watched stream and writing to the shared out.

        Communicates with the above methods via self.q. Runs until .shutdown()
        is called.
        """

        class Muxer:
            def __init__(self, q, out):
                self.q = q
                self.out = out
                self.paths = {}  # map path -> (f, decorator)
                self.running = False
                self.loop = asyncio.get_running_loop()

            def _do_read(self, f, decorator):
                while True:
                    line = f.readline()
                    if not line:
                        break
                    self.out.write(decorator(line))

            def watch(self, path, decorator=None):
                logger.debug(f'Watching {path}')
                if decorator is None:

                    def identity(s):
                        return s

                    decorator = identity
                assert path not in self.paths
                f = open(
                    os.open(str(path), os.O_RDONLY | os.O_NONBLOCK),
                    mode='r',
                    errors='surrogateescape',
                )
                self.loop.add_reader(f, self._do_read, f, decorator)
                self.paths[path] = f, decorator

            def unwatch(self, path):
                logger.debug(f'Unwatching {path}')
                assert path in self.paths
                f, decorator = self.paths.pop(path)
                self.loop.remove_reader(f)
                self._do_read(f, decorator)

            def shutdown(self):
                logger.debug('Shutting down')
                for path in sorted(self.paths.keys()):
                    logger.warning(f'{path} was not unwatched!')
                    self.unwatch(path)
                self.out.flush()
                self.running = False

            async def run(self):
                self.running = True
                while self.running:
                    cmd, *args = await self.q.get()
                    assert cmd in {'watch', 'unwatch', 'shutdown'}
                    getattr(self, cmd)(*args)
                    self.q.task_done()

        await Muxer(self.q, self.out).run()

    async def __aenter__(self):
        assert self._task is None
        self._task = asyncio.create_task(self.service())
        return self

    async def __aexit__(self, *_):
        assert self._task is not None
        assert not self._task.done()
        await self.shutdown()
        await self._task
        self._task = None



import asyncio
import concurrent.futures
import logging
import time

logger = logging.getLogger(__name__)


class Job:
    """Base class for jobs. Jobs have a name, dependencies, and are callable.

    A Job instance represents a unit of work to be performed. The job is
    identified by a name (.name), and its work is performed by calling it
    (i.e. invoking the .__call__() async method, aka. coroutine).

    Additionally, the job has a set of zero of more dependencies (.deps) which
    are the names of other jobs that should complete successfully before this
    job can proceed.

    A job instance is submitted for execution by passing it to the .add()
    method of a Scheduler instance. The Scheduler will then start the job
    (during its .run()) by starting the job's .__call__() coroutine.
    The scheduler passes itself to the job's .__call__(), to allow the job
    implementation to communicate with the scheduler while it's working.

    The return value from .__call__() is known as the job result. Any return
    value from .__call__() is taken to mean that the job finished successfully.
    This return value is accessible to other jobs (via Scheduler.results())
    and ultimately included in the result from Scheduler.run().

    If .__call__() raises an exception, the job is considered to have failed.
    This may or may not cause the Scheduler to cancel/abort all other jobs,
    but it will at least cause dependent jobs to be cancelled (when they call
    Scheduler.results()).

    The ordering of jobs according to their dependencies is implemented by the
    default .__call__() implementation below, by asking (and waiting for) the
    scheduler to return the results of its dependencies. Subclasses must make
    sure to call this base class implementation before assuming anything about
    the state of its dependencies.
    """

    def __init__(self, name, *, deps=None):
        self.name = name
        self.deps = set() if deps is None else set(deps)
        self.logger = logging.getLogger(name)
        self.logger.debug(f'Created {self.__class__.__name__}: {self}')

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name!r}, deps={self.deps!r})'

    async def __call__(self, scheduler):
        """Perform the job.

        The given scheduler object can be used to interact with the
        surrounding job running machinery.

        This superclass implementation awaits the results from our dependencies
        and returns their results as a dict mapping job names to results.
        Until this returns, we cannot assume that our dependencies have been
        run at all.

        Raises KeyError if any dependency has not been added to the scheduler.
        Raises CancelledError if any dependency has failed.
        """
        if self.deps:
            self.logger.debug(f'Awaiting dependencies {self.deps}…')
            return await scheduler.results(*self.deps)
        return {}


class Scheduler:
    """Async job scheduler. Run Job instances as concurrent tasks.

    Jobs are added by passing Job instances to .add(). These will be scheduled
    for execution when .run() is called. While running, jobs may await each
    other's results (by calling .results()). Further jobs may be added (with
    .add()). When all jobs are completed (with or without success), .run() will
    return dictionary with all the job results.
    """

    def __init__(self, *, event_handler=None):
        self.jobs = {}  # name -> Job
        self.tasks = {}  # name -> Task object, aka. (future) result from job()
        self.running = False
        self.event_handler = event_handler

    def __contains__(self, job_name):
        return job_name in self.jobs

    def event(self, event, job_name=None, info=None):
        if self.event_handler is None:
            return
        d = {'timestamp': time.time(), 'event': event}
        if job_name is not None:
            d['job'] = job_name
        if info is not None:
            d.update(info)
        logger.debug(f'Posting event: {d}')
        try:
            self.event_handler(d)
        except asyncio.QueueFull:
            logger.error('Failed to post event: {d}')

    @staticmethod
    def _fate(future):
        """Return a word describing the state of the given future."""
        if not future.done():
            return 'unfinished'
        elif future.cancelled():
            return 'cancelled'
        elif future.exception() is not None:
            return 'failed'
        else:
            return 'success'

    def _start_job(self, job):
        self.tasks[job.name] = asyncio.create_task(job(self), name=job.name)
        self.event('start', job.name)
        self.tasks[job.name].add_done_callback(
            lambda task: self.event(
                'finish', job.name, {'fate': self._fate(task)}
            )
        )

    def add(self, job):
        """Add a job to be run.

        If we're already running (i.e. inside .run()) schedule the job
        immediately, otherwise the job will be scheduled when .run() is called.
        Raise ValueError if a job with the same name has already been added.
        """
        if job.name in self.jobs:
            exist = self.jobs[job.name]
            raise ValueError(f'Cannot add {job} with same name as {exist}')
        self.jobs[job.name] = job
        self.event('add', job.name)
        if self.running:
            self._start_job(job)

    async def results(self, *job_names):
        """Wait until the given jobs are finished, and return their results.

        Returns a dictionary mapping job names to their results.

        Raise KeyError if any of the given jobs have not been added to the
        scheduler. If any of the given jobs raise an exception, then raise
        CancelledError here to cancel the calling job.
        """
        assert self.running
        caller = asyncio.current_task().get_name()
        tasks = [self.tasks[n] for n in job_names]
        pending = [n for n, t in zip(job_names, tasks) if not t.done()]
        self.event(
            'await results',
            caller,
            {'jobs': list(job_names), 'pending': pending},
        )
        if pending:
            logger.debug(f'{caller} is waiting for {", ".join(pending)}…')
        try:
            results = dict(zip(job_names, await asyncio.gather(*tasks)))
        except Exception:
            logger.info(f'{caller} cancelled due to failure(s) in {job_names}')
            raise asyncio.CancelledError
        self.event('awaited results', caller)
        logger.debug(f'{caller} <- {results} from .results()')
        return results

    async def _run_tasks(self, return_when):
        logger.info('Waiting for all jobs to complete…')
        self.event('await tasks', info={'jobs': list(self.tasks.keys())})
        try:
            while True:
                # Await the tasks that are currently running. More tasks may be
                # spawned while we're waiting, and those are not awaited here.
                await asyncio.wait(
                    self.tasks.values(), return_when=return_when
                )
                # We know _some_ task completed, successfully or not
                if return_when == concurrent.futures.FIRST_COMPLETED:
                    break
                elif return_when == concurrent.futures.FIRST_EXCEPTION:
                    # Figure out if there are failed tasks and we should return
                    if any(
                        t.done() and (t.cancelled() or t.exception())
                        for t in self.tasks.values()
                    ):
                        break
                # Otherwise return when all tasks are done
                if all(t.done() for t in self.tasks.values()):
                    break
        finally:
            # Any tasks left running at this point should be cancelled and
            # reaped/awaited _before_ we return from here
            for job_name, task in self.tasks.items():
                if not task.done():
                    logger.warning(f'Cancelling {job_name}...')
                    task.cancel()
            try:
                await asyncio.wait(self.tasks.values())
            finally:
                self.event('awaited tasks')

    async def run(self, keep_going=False):
        """Run until all jobs are finished.

        If keep_going is disabled (the default), the first failing job (i.e.
        a job that raises an exception) will cause us to cancel all other
        concurrent and remaining jobs and return as soon as possible.

        If keep_going is enabled, we will keep running as long as there are
        jobs to do. Only the jobs that depend (directly or indirectly) on the
        failing job(s) will be cancelled.

        Return a dictionary mapping job names to the corresponding
        asyncio.Future objects that encapsulate their fate (result, exception,
        or cancelled state).
        """
        logger.debug('Running…')
        self.event(
            'start',
            info={'keep_going': keep_going, 'num_jobs': len(self.jobs)},
        )
        if keep_going:
            return_when = concurrent.futures.ALL_COMPLETED
        else:
            return_when = concurrent.futures.FIRST_EXCEPTION

        to_start = list(self.jobs.values())
        if to_start:
            for job in to_start:
                self._start_job(job)

            self.running = True
            await self._run_tasks(return_when)
            self.running = False
        else:
            logger.warning('Nothing to do!')
        self.event('finish', info={'num_tasks': len(self.tasks)})
        return self.tasks



"""Add missing async functionality to Python v3.6 and v3.7."""
import asyncio
import contextlib


# asyncio.Task names added in Python v3.8
if not hasattr(asyncio.Task, 'get_name'):

    class Task(asyncio.Task):
        num_tasks = 0

        def __init__(self, coro, *, loop=None, name=None):
            super().__init__(coro, loop=loop)

            self.__class__.num_tasks += 1
            if name is None:
                self._name = f'Task-{self.__class__.num_tasks}'
            else:
                self._name = str(name)

        def get_name(self):
            return self._name

        def set_name(self, value):
            self._name = str(value)

    def create_task(coro, *, name=None):
        return asyncio.ensure_future(Task(coro, name=name))

    setattr(asyncio, 'Task', Task)
    setattr(asyncio, 'create_task', create_task)


# asyncio.current_task() added in Python v3.7
if not hasattr(asyncio, 'current_task'):

    def current_task():
        return asyncio.Task.current_task()

    setattr(asyncio, 'current_task', current_task)


# asyncio.get_running_loop() added in Python v3.7
if not hasattr(asyncio, 'get_running_loop'):

    def get_running_loop():
        return asyncio.get_event_loop()

    setattr(asyncio, 'get_running_loop', get_running_loop)


# asyncio.run() added in Python v3.7
if not hasattr(asyncio, 'run'):

    def run(main, *, debug=False):
        loop = asyncio.get_event_loop()
        ret = loop.run_until_complete(main)
        loop.close()
        return ret

    setattr(asyncio, 'run', run)


# contextlib.asynccontextmanager added in Python v3.7
if not hasattr(contextlib, 'asynccontextmanager'):
    # The following was pilfered and simplified from Python v3.8's contextlib:
    # https://github.com/python/cpython/blob/3.8/Lib/contextlib.py
    from functools import wraps

    class _AsyncGeneratorContextManager:
        def __init__(self, func, args, kwds):
            self.gen = func(*args, **kwds)
            self.func, self.args, self.kwds = func, args, kwds

        async def __aenter__(self):
            try:
                return await self.gen.__anext__()
            except StopAsyncIteration:
                raise RuntimeError("generator didn't yield") from None

        async def __aexit__(self, typ, value, traceback):
            if typ is None:
                try:
                    await self.gen.__anext__()
                except StopAsyncIteration:
                    return
                else:
                    raise RuntimeError("generator didn't stop")
            else:
                if value is None:
                    value = typ()
                # See _GeneratorContextManager.__exit__ for comments on
                # subtleties in this implementation
                try:
                    await self.gen.athrow(typ, value, traceback)
                    raise RuntimeError("generator didn't stop after athrow()")
                except StopAsyncIteration as exc:
                    return exc is not value
                except RuntimeError as exc:
                    if exc is value:
                        return False
                    # Avoid suppressing if a StopIteration exception
                    # was passed to throw() and later wrapped into a
                    #  RuntimeError (see PEP 479 for sync generators; async
                    # generators also have this behavior). But do this only if
                    # the exception wrapped by the RuntimeError is actully
                    # Stop(Async)Iteration (see issue29692).
                    if isinstance(value, (StopIteration, StopAsyncIteration)):
                        if exc.__cause__ is value:
                            return False
                    raise
                except BaseException as exc:
                    if exc is not value:
                        raise

    def asynccontextmanager(func):
        @wraps(func)
        def helper(*args, **kwds):
            return _AsyncGeneratorContextManager(func, args, kwds)

        return helper

    setattr(contextlib, 'asynccontextmanager', asynccontextmanager)



import logging
import pytest
from subprocess import CalledProcessError

from asyncjobs import external_work

from conftest import (
    Cancelled,
    setup_scheduler,
    TExternalWorkJob,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

logger = logging.getLogger(__name__)


@pytest.fixture
def run(scheduler_with_workers):
    Scheduler = scheduler_with_workers(external_work.Scheduler)

    async def _run(todo, **run_args):
        with setup_scheduler(Scheduler, todo) as scheduler:
            return await scheduler.run(**run_args)

    return _run


TJob = TExternalWorkJob


async def test_one_ok_job_in_thread(run):
    todo = [TJob('foo', thread=lambda: 'foo worked')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo worked'})


async def test_one_ok_job_in_subproc(run, tmp_path):
    path = tmp_path / 'foo'
    todo = [TJob('foo', subproc=['touch', str(path)])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 0})
    assert path.is_file()


async def test_one_failed_between_two_ok_jobs_in_threads_cancels_last(run):
    def raiseUGH():
        raise ValueError('UGH')

    todo = [
        TJob('foo', before={'bar'}, thread=lambda: 'foo worked'),
        TJob('bar', {'foo'}, before={'baz'}, thread=raiseUGH),
        TJob('baz', {'bar'}, thread=lambda: 'baz worked'),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo worked', 'bar': ValueError('UGH'), 'baz': Cancelled}
    )


async def test_one_failed_between_two_in_subprocs_cancels_last(run, tmp_path):
    foo_path = tmp_path / 'foo'
    baz_path = tmp_path / 'baz'
    todo = [
        TJob('foo', before={'bar'}, subproc=['touch', str(foo_path)]),
        TJob('bar', {'foo'}, before={'baz'}, subproc=['false']),
        TJob('baz', {'bar'}, subproc=['touch', str(baz_path)]),
    ]
    done = await run(todo)
    assert verify_tasks(
        done,
        {'foo': 0, 'bar': CalledProcessError(1, ['false']), 'baz': Cancelled},
    )
    assert foo_path.is_file()
    assert not baz_path.exists()



import itertools

import plotly.colors as plotly_colors
import plotly.graph_objects as go


def plottable_events(events):
    epoch = None
    for e in events:
        if epoch is None:
            epoch = e['timestamp']
        assert e['timestamp'] >= epoch
        wait_for = []
        if e['event'] == 'await results':
            wait_for = sorted(e['jobs'])
        details = ', '.join(wait_for) if wait_for else ''
        if 'job' not in e:
            continue
        yield dict(
            event=e['event'],
            task=e['job'],
            time=e['timestamp'] - epoch,
            done=e['event'] == 'finish',
            label=e['event'] + (f' ({details})' if details else ''),
            wait_for=wait_for,
        )


def plot_schedule(
    title,
    events,
    colors=None,
    bar_width=0.45,
    showgrid_x=True,
    showgrid_y=True,
    **kwargs,
):
    fig = go.Figure(
        layout=dict(
            title=title,
            showlegend=False,
            hovermode='closest',
            xaxis=dict(
                title='Time [seconds]',
                autorange=True,
                showgrid=showgrid_x,
                zeroline=False,
            ),
            yaxis=dict(
                title='Jobs',
                showgrid=showgrid_y,
                ticktext=[],  # rows.keys()
                tickvals=[],  # rows.values()
                range=[-1, 1],  # use rows.values() to update
                autorange=False,
                zeroline=False,
            ),
            **kwargs,
        ),
    )

    # Events are grouped into rows by their 'task'. The tasks are ordered
    # vertically by the appearance of its first event. Rows are drawn top-down,
    # i.e. in the _negative_ y-direction.
    rows = {}  # map task name -> y-coordinate from 0 and down

    def get_row(task):
        if task not in rows:
            rows[task] = -len(rows)
            fig.update_layout(
                dict(
                    yaxis=dict(
                        ticktext=list(rows.keys()),
                        tickvals=list(rows.values()),
                        range=[min(rows.values()) - 1, max(rows.values()) + 1],
                    )
                )
            )
        return rows[task]

    if colors is None:
        colors = {}  # map task name -> color
    # colors not already in colors are fetched from the list of default colors:
    palette = itertools.cycle(plotly_colors.DEFAULT_PLOTLY_COLORS)

    def get_color(task):
        if task not in colors:
            colors[task] = next(palette)
        return colors[task]

    def marker(event, y, color):
        """Draw a marker for the given event."""
        symbol = {
            'add': 'star',
            'start': 'diamond-wide-open',
            'await results': 'diamond-tall-open',
            'awaited results': 'diamond-wide-open',
            'await worker slot': 'diamond-tall-open',
            'awaited worker slot': 'diamond-wide-open',
            'await worker thread': 'triangle-right',
            'awaited worker thread': 'triangle-left',
            'await worker proc': 'triangle-right',
            'awaited worker proc': 'triangle-left',
            'finish': 'x',
        }
        return dict(
            name='',  # event['task'],
            legendgroup=event['task'],
            mode='markers',
            # ids=[...]
            x=[event['time']],
            y=[y],
            text=[event['label']],
            marker=dict(
                symbol=symbol.get(event['event'], 'square'),
                size=10,
                color=color,
            ),
        )

    def rect(a, b, y, color):
        """Draw a rectangle between the two given events.

        The label/metadata of the rectangle are taken from the first event.
        """
        assert a['task'] == b['task']
        x0, x1 = a['time'], b['time']
        y0, y1 = y - bar_width, y + bar_width
        opacity = {
            'add': 0.2,
            'start': 0.6,
            'await results': 0.2,
            'awaited results': 0.6,
            'await worker slot': 0.2,
            'awaited worker slot': 0.6,
            'await worker thread': 0.8,
            'awaited worker thread': 0.6,
            'await worker proc': 0.8,
            'awaited worker proc': 0.6,
        }
        return dict(
            name=a['task'],
            legendgroup=a['task'],
            opacity=opacity.get(a['event'], 1),
            mode='none',
            # ids=[...]
            # 4 corners, clockwise from top-left
            x=[x0, x1, x1, x0],
            y=[y1, y1, y0, y0],
            fill='toself',  # fill area enclosed by above points
            fillcolor=color,
            hoverinfo='text',
            text=a['label'],
        )

    def line(a, ay, b, by, color):
        """Draw a line from event a at y == ay to event b at y == by."""
        return dict(
            name='',
            opacity=0.75,
            mode='lines',
            x=[b['time'], a['time']],
            y=[by, ay],
            text=f"{a['task']} -> {b['task']}",
            line=dict(color=color, width=2, shape='hv', dash='dot'),
        )

    last_event_by_task = {}  # map task name -> last event seen for that task
    for e in plottable_events(events):
        task = e['task']
        y = get_row(task)
        color = get_color(task)
        fig.add_trace(marker(e, y, color))
        if task in last_event_by_task:
            last = last_event_by_task[task]
            fig.add_trace(rect(last, e, y, color))
            for t in last['wait_for']:
                other = last_event_by_task[t]
                assert other['done']
                fig.add_trace(line(other, get_row(t), e, y, color))
        last_event_by_task[task] = e

    return fig



from pathlib import Path
import pytest
import subprocess

example_scripts = (Path(__file__).parent.parent / 'examples').glob('*.py')


@pytest.mark.parametrize('script', example_scripts)
def test_example_script(script):
    subprocess.run([script], check=True)



#!/usr/bin/env python3
"""Create random jobs that sleep in threads, and plot their scheduling."""

import argparse
import asyncio
import itertools
import logging
import random
import sys
import time

from asyncjobs import Job, Scheduler

logger = logging.getLogger('random_jobs')


class TimeWaster(Job):
    def __init__(self, work, **kwargs):
        self.work = work
        super().__init__(**kwargs)
        self.logger.info(self)

    def __str__(self):
        deps = ', '.join(sorted(self.deps))
        return f'{super().__str__()}[{deps}]/{self.work}'

    def do_work(self):
        self.logger.info(f'Doing own work: {self.work}')
        time.sleep(self.work / 1000)
        self.logger.info(f'Finished own work: {self.work}')
        return {self.name: self.work}

    async def __call__(self, scheduler):
        self.dep_results = await super().__call__(scheduler)
        return await self.call_in_thread(self.do_work)


class ParallelTimeWaster(TimeWaster):
    def __init__(self, *, work_threshold, **kwargs):
        self.work_threshold = work_threshold
        super().__init__(**kwargs)

    async def __call__(self, scheduler):
        i = 0
        while self.work > self.work_threshold:
            i += 1
            work = random.randint(
                self.work_threshold * 2 // 3, self.work_threshold
            )
            name = f'{self.name}_{i}/{work}'
            self.logger.info(f'Splitting off {name}')
            scheduler.add(TimeWaster(name=name, work=work))
            self.deps.add(name)
            self.work -= work
        return await super().__call__(scheduler)

    def do_work(self):
        result = {}
        self.logger.info('From deps:')
        for dep, dep_result in self.dep_results.items():
            self.logger.info(f'  {dep}: {dep_result}')
            result.update(dep_result)
        result.update(super().do_work())
        self.logger.info(f'Returning {result}')
        return result


class RandomJob(ParallelTimeWaster):
    @classmethod
    def generate(cls, dep_prob, max_work=100, work_threshold=None):
        if work_threshold is None:
            work_threshold = max_work // 2
            assert work_threshold > 0
        names = []
        i = 0
        while True:
            letter = chr(ord('a') + i)
            work = random.randint(0, max_work)
            name = f'{letter}/{work}'
            yield cls(
                name=name,
                deps=set(filter(lambda _: random.random() < dep_prob, names)),
                work=work,
                work_threshold=work_threshold,
            )
            names.append(name)
            i += 1


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'num_jobs',
        type=int,
        nargs='?',
        default=10,
        help='Number of jobs to run',
    )
    parser.add_argument(
        'dep_prob',
        type=float,
        nargs='?',
        default=0.5,
        help='Probability of depending on each preceding job',
    )
    parser.add_argument(
        'max_work',
        type=int,
        nargs='?',
        default=100,
        help='Max duration of each job (msecs)',
    )
    parser.add_argument(
        'workers',
        type=int,
        nargs='?',
        default=4,
        help='Max number of parallel workers',
    )
    parser.add_argument(
        '-p', '--plot', action='store_true', help='Plot job schedule'
    )
    parser.add_argument(
        '-v', '--verbose', action='count', default=0, help='Increase log level'
    )
    parser.add_argument(
        '-q', '--quiet', action='count', default=0, help='Decrease log level'
    )
    args = parser.parse_args()

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(
        logging.Formatter(
            fmt=(
                '{relativeCreated:8.0f} {process:5}/{threadName:10} '
                '{name:>16}: {message}'
            ),
            style='{',
        )
    )
    loglevel = logging.WARNING + 10 * (args.quiet - args.verbose)
    logging.basicConfig(handlers=[handler], level=loglevel)

    print(f'Generating {args.num_jobs} jobs of work <= {args.max_work}…')
    random.seed(0)  # deterministic
    job_generator = RandomJob.generate(args.dep_prob, args.max_work)
    jobs = list(itertools.islice(job_generator, args.num_jobs))

    events = []
    builder = Scheduler(workers=args.workers, event_handler=events.append)
    for job in jobs:
        builder.add(job)
    results = asyncio.run(builder.run(), debug=False)
    longest_work = max(sum(f.result().values()) for f in results.values())
    print(f'Finished with max(sum(work)) == {longest_work}')

    if args.plot:
        from asyncjobs.plot_schedule import plot_schedule

        plot_schedule(title=' '.join(sys.argv), events=events).show()


if __name__ == '__main__':
    main()



import asyncio
from contextlib import contextmanager
from functools import partial
import logging
import signal

from . import basic

logger = logging.getLogger(__name__)


Job = basic.Job


class Scheduler(basic.Scheduler):
    """Teach Scheduler to cancel/abort properly on incoming signals.

    This installs appropriate signal handler to intercept SIGHUP, SIGTERM and
    SIGINT (aka. Ctrl+C or KeyboardInterrupt) and cause them to cancel and
    clean up the Scheduler (and its jobs) in an orderly manner.
    """

    handle_signals = {signal.SIGHUP, signal.SIGTERM, signal.SIGINT}

    def _caught_signal(self, signum, this_task):
        logger.warning(f'Caught signal {signum}!')
        assert self.running and not this_task.done()
        logger.warning(f'Cancelling task {this_task}…')
        this_task.cancel()

    @contextmanager
    def _handle_signals(self):
        loop = asyncio.get_running_loop()
        sched_task = asyncio.current_task()
        for signum in self.handle_signals:
            handler = partial(self._caught_signal, signum, sched_task)
            loop.add_signal_handler(
                signum, partial(loop.call_soon_threadsafe, handler)
            )
        try:
            yield
        except asyncio.CancelledError:
            logger.warning('Intercepted Cancelled on the way out')
        finally:
            for signum in self.handle_signals:
                loop.remove_signal_handler(signum)

    async def _run_tasks(self, *args, **kwargs):
        with self._handle_signals():
            return await super()._run_tasks(*args, **kwargs)



import logging
import pytest

from asyncjobs import external_work, signal_handling

from conftest import (
    abort_in,
    assert_elapsed_time_within,
    Cancelled,
    setup_scheduler,
    TExternalWorkJob,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

logger = logging.getLogger(__name__)


@pytest.fixture
def run(scheduler_with_workers):
    Scheduler = scheduler_with_workers(
        signal_handling.Scheduler, external_work.Scheduler
    )

    async def _run(todo, abort_after=None, **run_args):
        with setup_scheduler(Scheduler, todo) as scheduler:
            with abort_in(abort_after):
                return await scheduler.run(**run_args)

    return _run


TJob = TExternalWorkJob

# aborting jobs shall properly clean up all jobs + scheduler


async def test_return_before_abort(run):
    todo = [TJob('foo', async_sleep=0.1)]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.3)
    assert verify_tasks(done, {'foo': 'foo done'})


async def test_abort_one_job_returns_immediately(run):
    todo = [TJob('foo', async_sleep=0.3)]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_job_in_thread_returns_immediately(run):
    todo = [TJob('foo', thread_sleep=0.3)]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_job_in_subproc_returns_immediately(run):
    todo = [TJob('foo', subproc_sleep=30)]
    with assert_elapsed_time_within(0.3):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_spawned_job_returns_immediately(run):
    todo = [
        TJob('foo', spawn=[TJob('bar', async_sleep=0.3)], await_spawn=True)
    ]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled, 'bar': Cancelled})


async def test_abort_hundred_jobs_returns_immediately(run):
    todo = [TJob(f'foo #{i}', async_sleep=0.3) for i in range(100)]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {f'foo #{i}': Cancelled for i in range(100)})


async def test_abort_hundred_jobs_in_threads_returns_immediately(run):
    todo = [TJob(f'foo #{i}', thread_sleep=1.0) for i in range(100)]
    with assert_elapsed_time_within(0.5):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {f'foo #{i}': Cancelled for i in range(100)})


async def test_abort_hundred_jobs_in_subprocs_returns_immediately(run):
    todo = [TJob(f'foo #{i}', subproc_sleep=5.0) for i in range(100)]
    with assert_elapsed_time_within(2.0):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {f'foo #{i}': Cancelled for i in range(100)})


async def test_abort_hundred_spawned_jobs_returns_immediately(run):
    todo = [
        TJob(
            'foo',
            spawn=[TJob(f'bar #{i}', async_sleep=0.3) for i in range(100)],
            await_spawn=True,
        )
    ]
    with assert_elapsed_time_within(0.2):
        done = await run(todo, abort_after=0.1)
    expect = {f'bar #{i}': Cancelled for i in range(100)}
    expect['foo'] = Cancelled
    assert verify_tasks(done, expect)



import asyncio
import concurrent.futures
import contextlib
import logging
import subprocess
from typing import Any, Callable, List

from . import basic

logger = logging.getLogger(__name__)


class Job(basic.Job):
    # Override _one_ of these in a subclass or instance to have it
    # automatically invoked by the default __call__() implementation.
    # For anything more advanced, please override __call__() instead.
    thread_func: Callable[[], Any] = NotImplemented
    subprocess_argv: List[str] = NotImplemented

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scheduler = None

    async def __call__(self, scheduler):
        self._scheduler = scheduler
        ret = await super().__call__(scheduler)
        if self.thread_func is not NotImplemented:
            ret = await self.call_in_thread(self.thread_func)
        elif self.subprocess_argv is not NotImplemented:
            ret = await self.run_in_subprocess(self.subprocess_argv)
        return ret

    async def call_in_thread(self, func, *args):
        """Call func(*args) in a worker thread and await its result."""
        assert self._scheduler is not None
        return await self._scheduler.call_in_thread(func, *args)

    async def run_in_subprocess(self, argv, **kwargs):
        """Run a command line in a subprocess and await its exit code."""
        assert self._scheduler is not None
        return await self._scheduler.run_in_subprocess(argv, **kwargs)


class Scheduler(basic.Scheduler):
    """Manage jobs whose work is done in threads or subprocesses.

    Extend Scheduler with methods that allow Job instances to perform work in
    other threads or subprocess, while keeping the number of _concurrent_
    threads/processes within the given limit.
    """

    def __init__(self, *, workers=1, **kwargs):
        assert workers > 0
        self.workers = workers
        self.worker_sem = None
        self.worker_threads = None
        super().__init__(**kwargs)

    @contextlib.asynccontextmanager
    async def reserve_worker(self, caller=None):
        """Acquire a worker context where the caller can run its own work.

        This is the mechanism that ensures we do not schedule more concurrent
        work than allowed by .workers. Anybody that wants to spin off another
        thread or process to perform some work should use this context manager
        to await a free "slot".
        """
        assert self.running
        logger.debug('acquiring worker semaphore…')
        self.event('await worker slot', caller)
        async with self.worker_sem:
            self.event('awaited worker slot', caller)
            logger.debug('acquired worker semaphore')
            yield
            logger.debug('releasing worker semaphore')

    async def call_in_thread(self, func, *args):
        """Call func(*args) in a worker thread and await its result."""
        caller = asyncio.current_task().get_name()
        async with self.reserve_worker(caller):
            if self.worker_threads is None:
                self.worker_threads = concurrent.futures.ThreadPoolExecutor(
                    max_workers=self.workers,
                    thread_name_prefix='Worker Thread',
                )
            try:
                logger.debug(f'{caller} -> awaiting worker thread…')
                self.event('await worker thread', caller, {'func': str(func)})
                future = asyncio.get_running_loop().run_in_executor(
                    self.worker_threads, func, *args
                )
                future.add_done_callback(
                    lambda f: self.event(
                        'awaited worker thread',
                        caller,
                        {'fate': self._fate(f)},
                    )
                )
                result = await future
                logger.debug(f'{caller} <- {result!r} from worker')
                return result
            except Exception as e:
                logger.warning(f'{caller} <- Exception {e} from worker!')
                raise

    async def run_in_subprocess(
        self, argv, stdin=None, stdout=None, stderr=None, check=True
    ):
        """Run a command line in a subprocess and await its exit code."""
        caller = asyncio.current_task().get_name()
        retcode = None
        async with self.reserve_worker(caller):
            logger.debug(f'{caller} -> starting {argv} in subprocess…')
            self.event('await worker proc', caller, {'argv': argv})
            proc = await asyncio.create_subprocess_exec(
                *argv, stdin=stdin, stdout=stdout, stderr=stderr
            )
            try:
                logger.debug(f'{caller} -- awaiting subprocess…')
                retcode = await proc.wait()
            except asyncio.CancelledError:
                logger.error(f'{caller} cancelled! Terminating {proc}!…')
                proc.terminate()
                try:
                    retcode = await proc.wait()
                    logger.debug(f'{caller} cancelled! {proc} terminated.')
                except asyncio.CancelledError:
                    logger.error(f'{caller} cancelled again! Killing {proc}!…')
                    proc.kill()
                    retcode = await proc.wait()
                    logger.debug(f'{caller} cancelled! {proc} killed.')
                raise
            finally:
                self.event('awaited worker proc', caller, {'exit': retcode})
        if check and retcode != 0:
            raise subprocess.CalledProcessError(retcode, argv)
        return retcode

    async def _run_tasks(self, *args, **kwargs):
        self.worker_sem = asyncio.BoundedSemaphore(self.workers)

        try:
            await super()._run_tasks(*args, **kwargs)
            if self.worker_threads is not None:
                logger.debug('Shutting down worker threads…')
                self.worker_threads.shutdown()  # wait=timeout is None)
                self.worker_threads = None
                logger.debug('Shut down worker threads')
        finally:
            if self.worker_threads is not None:
                logger.error('Cancelling without awaiting workers…')
                self.worker_threads.shutdown(wait=False)
                self.worker_threads = None



from . import polyfill  # noqa: F401
from . import external_work, signal_handling


class Job(external_work.Job, signal_handling.Job):
    pass


class Scheduler(external_work.Scheduler, signal_handling.Scheduler):
    pass



import asyncio
import logging
from pathlib import Path
import pytest
import random
import time

from asyncjobs import logmuxed_work, signal_handling

from conftest import (
    abort_in,
    assert_elapsed_time_within,
    setup_scheduler,
    TExternalWorkJob,
)

pytestmark = pytest.mark.asyncio

logger = logging.getLogger(__name__)

sh_helper = Path(__file__).parent / 'test_5_logmuxed_work_helper.sh'


class TJob(logmuxed_work.Job, TExternalWorkJob):
    def __init__(
        self,
        *args,
        out=None,
        err=None,
        in_thread=False,
        decorate=False,
        redirect_logger=False,
        **kwargs,
    ):
        super().__init__(*args, redirect_logger=redirect_logger, **kwargs)
        # out/err are strings or lists of strings. If strings, the chars in the
        # string will be printed one-by-one, if lists of strings, each string
        # will be printed. See ._print_out_and_err() below for details.
        self.out = out
        self.err = err
        self.decorate = decorate
        if in_thread:
            self.thread = self._print_out_and_err(True)
        elif 'subproc' not in kwargs:
            self.do_work = self._print_out_and_err(False)

    def decorate_out(self, msg):
        if self.decorate:
            return f'{self.name}/out: {msg.rstrip()}\n'
        else:
            return msg

    def decorate_err(self, msg):
        if self.decorate:
            return f'{self.name}/ERR: {msg.rstrip()}\n'
        else:
            return msg

    def _print_out_and_err(self, sync):
        # Try really hard to provoke any issues we might have regarding
        # rescheduling/ordering of tasks and whatnot: Return a sync/async
        # callable that prints one item (char or string) from either self.out
        # or self.err (to self.stdout/stderr respectively), and then yields/
        # sleeps to allow other things to run before the next item is printed.
        # Keep going until self.out and self.err are exhausted.

        if isinstance(self.out, list):  # list of strings
            out = [line + '\n' for line in self.out]
        else:  # list of chars
            out = [] if self.out is None else list(self.out)
        if isinstance(self.err, list):  # list of strings
            err = [line + '\n' for line in self.err]
        else:  # list of chars
            err = [] if self.err is None else list(self.err)

        def print_one_item():
            if not (out or err):  # Nothing left to do
                return False
            if out and err:  # Choose one at random
                src, dst = random.choice(
                    [(out, self.stdout), (err, self.stderr)]
                )
            elif out:
                src, dst = out, self.stdout
            else:
                src, dst = err, self.stderr
            assert src
            print(src.pop(0), file=dst, end='')
            return True

        if sync:

            def print_sync():
                while True:
                    if not print_one_item():
                        break
                    time.sleep(0.001)

            return print_sync
        else:

            async def print_async(*_):
                while True:
                    if not print_one_item():
                        break
                    await asyncio.sleep(0)

            return print_async


@pytest.fixture
def run(scheduler_with_workers):
    Scheduler = scheduler_with_workers(
        signal_handling.Scheduler, logmuxed_work.Scheduler
    )

    async def _run(todo, abort_after=None, **run_args):
        with setup_scheduler(Scheduler, todo) as scheduler:
            with abort_in(abort_after):
                return await scheduler.run(**run_args)

    return _run


async def test_no_output_from_no_jobs(run, verify_output):
    await run([])
    assert verify_output([], [])


async def test_no_output_from_two_jobs(run, verify_output):
    await run([TJob('foo'), TJob('bar')])
    assert verify_output([[], []], [[], []])


async def test_default_output_is_undecorated(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        # Pass out/err as lists of strings to print entire strings at once
        TJob(name, out=[out.format(name=name)], err=[err.format(name=name)])
        for name in jobs
    ]
    await run(todo)
    assert verify_output(
        [[out.format(name=name)] for name in jobs],
        [[err.format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_two_jobs(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            out=out.format(name=name),
            err=err.format(name=name),
            decorate=True,
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_spawned_jobs(run, verify_output):
    foo = TJob(
        'foo',
        out="This is foo's stdout",
        err="This is foo's stderr",
        decorate=True,
        spawn=[
            TJob(
                'bar',
                out="This is bar's stdout",
                err="This is bar's stderr",
                decorate=True,
                spawn=[
                    TJob(
                        'baz',
                        out="This is baz's stdout",
                        err="This is baz's stderr",
                        decorate=True,
                    )
                ],
            )
        ],
    )
    await run([foo])
    assert verify_output(
        [
            ["foo/out: This is foo's stdout"],
            ["bar/out: This is bar's stdout"],
            ["baz/out: This is baz's stdout"],
        ],
        [
            ["foo/ERR: This is foo's stderr"],
            ["bar/ERR: This is bar's stderr"],
            ["baz/ERR: This is baz's stderr"],
        ],
    )


async def test_decorated_output_from_thread_worker(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            out=out.format(name=name),
            err=err.format(name=name),
            decorate=True,
            in_thread=True,
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_subprocess_worker(run, verify_output):
    assert sh_helper.is_file()
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            decorate=True,
            subproc=[
                str(sh_helper),
                out.format(name=name),
                err.format(name=name),
            ],
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_aborted_processes(
    num_workers, run, verify_output
):
    assert sh_helper.is_file()
    jobs = ['foo', 'bar', 'baz']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            decorate=True,
            subproc=[
                str(sh_helper),
                out.format(name=name),
                err.format(name=name),
                '1',  # sleep 1 sec between stdout print and stderr print
            ],
        )
        for name in jobs
    ]
    with assert_elapsed_time_within(0.75):
        await run(todo, abort_after=0.5)

    # We have 3 jobs, but can only run as many concurrently as there are
    # workers available. The rest will be cancelled before they start their
    # subprocee
    outprefix = '{name}/out: '
    expect_out = [[(outprefix + out).format(name=name)] for name in jobs]
    assert verify_output(expect_out[:num_workers], [])


async def test_decorated_output_from_100_jobs(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            out=out.format(i=i),
            err=err.format(i=i),
            decorate=True,
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_spawned_jobs(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = TJob(
        'foo',
        spawn=[
            TJob(
                f'job #{i}',
                out=out.format(i=i),
                err=err.format(i=i),
                decorate=True,
            )
            for i in range(100)
        ],
    )
    await run([todo])

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_thread_workers(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            out=out.format(i=i),
            err=err.format(i=i),
            decorate=True,
            in_thread=True,
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_subprocesses(run, verify_output):
    assert sh_helper.is_file()
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            decorate=True,
            subproc=[str(sh_helper), out.format(i=i), err.format(i=i)],
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )



from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="asyncjobs",
    version="0.2.0",
    description="Asynchronous job scheduler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jherland/asyncjobs",
    author="Johan Herland",
    author_email="johan@herland.net",
    packages=["asyncjobs"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    keywords="async asyncio job scheduler",
    python_requires=">=3.6",
    install_requires=[],
    extras_require={
        "dev": ["black", "flake8", "nox"],
        "dist": ["check-manifest", "twine", "wheel"],
        "plot": ["numpy", "plotly"],
        "test": ["pytest>=5.4.0", "pytest-asyncio", "requests"],
    },
)



import nox

# Run everything but 'dist' by default
nox.options.keywords = "not dist"


@nox.session(python=["3.6", "3.7", "3.8", "3.9"])
def tests(session):
    session.install(".[test]")
    session.run("pytest", "-x", "--log-level=debug")


@nox.session
def format(session):
    session.install("black")
    session.run(
        "black",
        "--target-version=py36",
        "--line-length=79",
        "--skip-string-normalization",
        ".",
    )


@nox.session
def lint(session):
    session.install("flake8")
    session.run("flake8")


@nox.session
def dist(session):
    session.install(".[dist]")
    session.run("check-manifest")
    session.run("python", "setup.py", "bdist_wheel", "sdist")
    session.run("twine", "upload", "dist/*")
    print("*** Don't forget to tag and push!")



import pytest

from asyncjobs.basic import Scheduler

from conftest import (
    Cancelled,
    setup_scheduler,
    TSimpleJob,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio


async def run(todo, **run_args):
    with setup_scheduler(Scheduler, todo) as scheduler:
        return await scheduler.run(**run_args)


TJob = TSimpleJob


async def test_zero_jobs_does_nothing():
    assert await run([]) == {}


async def test_one_ok_job():
    todo = [TJob('foo')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done'})


async def test_one_failed_job():
    done = await run([TJob('foo', result=ValueError('UGH'))])
    assert len(done) == 1
    e = done['foo'].exception()
    assert isinstance(e, ValueError) and e.args == ('UGH',)
    with pytest.raises(ValueError, match='UGH'):
        done['foo'].result()


async def test_cannot_add_second_job_with_same_name():
    with pytest.raises(ValueError):
        await run([TJob('foo'), TJob('foo')])


async def test_job_with_nonexisting_dependency_raises_KeyError():
    done = await run([TJob('foo', {'MISSING'})])
    with pytest.raises(KeyError, match='MISSING'):
        done['foo'].result()


async def test_two_independent_ok_jobs():
    todo = [TJob('foo'), TJob('bar')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


async def test_one_ok_before_another_ok_job():
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


async def test_one_ok_before_one_failed_job():
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}, result=ValueError('UGH')),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': ValueError('UGH')})


async def test_one_failed_job_before_one_ok_job_cancels_second_job():
    todo = [
        TJob('foo', before={'bar'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': Cancelled})


async def test_one_failed_job_before_two_ok_jobs_cancels_two_jobs():
    todo = [
        TJob('foo', before={'bar', 'baz'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}),
        TJob('baz', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': ValueError('UGH'), 'bar': Cancelled, 'baz': Cancelled}
    )


async def test_one_failed_job_before_two_dependent_jobs_cancels_two_jobs():
    todo = [
        TJob('foo', before={'bar'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}, before={'baz'}),
        TJob('baz', {'bar'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': ValueError('UGH'), 'bar': Cancelled, 'baz': Cancelled}
    )


async def test_one_failed_job_between_two_ok_jobs_cancels_last_job():
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}, before={'baz'}, result=ValueError('UGH')),
        TJob('baz', {'bar'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo done', 'bar': ValueError('UGH'), 'baz': Cancelled}
    )


async def test_one_ok_and_one_failed_job_without_keep_going_cancels_ok_job():
    todo = [
        TJob('foo', result=ValueError('UGH')),
        TJob('bar', async_sleep=0.01),  # allow time for potential cancellation
    ]
    done = await run(todo, keep_going=False)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': Cancelled})


async def test_one_ok_and_one_failed_job_with_keep_going_runs_ok_job():
    todo = [
        TJob('foo', result=ValueError('UGH')),
        TJob('bar', async_sleep=0.01),  # allow time for potential cancellation
    ]
    done = await run(todo, keep_going=True)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': 'bar done'})


# jobs that spawn child jobs


async def test_one_job_spawns_another():
    todo = [TJob('foo', spawn=[TJob('bar')])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


async def test_one_job_spawns_two_with_deps():
    # foo start bar and baz, baz depends on bar, foo waits for both to finish
    todo = [
        TJob(
            'foo',
            spawn=[
                TJob('bar', before={'foo', 'baz'}),
                TJob('baz', {'bar'}, before={'foo'}),
            ],
            await_spawn=True,
        )
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo done', 'bar': 'bar done', 'baz': 'baz done'}
    )


async def test_one_job_spawns_failing_job():
    todo = [TJob('foo', spawn=[TJob('bar', result=ValueError('UGH'))])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': ValueError('UGH')})


async def test_job_is_cancelled_when_waiting_for_failing_spawn():
    todo = [
        TJob(
            'foo',
            spawn=[TJob('bar', result=ValueError('UGH'))],
            await_spawn=True,
        )
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': Cancelled, 'bar': ValueError('UGH')})


async def test_spawn_outliving_parent_is_not_cancelled_by_scheduler():
    todo = [TJob('foo', spawn=[TJob('bar', async_sleep=0.01)])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})



#!/usr/bin/env python3
# Fetch a random wikipedia article and follow links from it
import asyncio
from asyncjobs import Job, Scheduler
import requests
import sys
import xml.etree.ElementTree as ET

num_articles = 10
num_workers = 4


class Article:
    base_url = 'https://en.wikipedia.org'
    random_url = base_url + '/wiki/Special:Random'

    def __init__(self, url):
        resp = requests.get(url=url, allow_redirects=True)
        self.html = ET.fromstring(resp.content)

    def title(self):
        ret = self.html.find('head').find('title').text
        if ret.endswith('- Wikipedia'):
            ret = ret.rsplit('-', 1)[0].strip()
        return ret

    def hrefs(self):
        for a in self.html.iterfind('.//p//a'):
            href = a.attrib.get('href', '')
            if href.startswith('/wiki/'):
                yield self.base_url + href


class Fetcher(Job):
    def __init__(self, url, level=1, **kwargs):
        super().__init__(name=url, **kwargs)
        self.url = url
        self.level = level

    def thread_func(self):
        indent = '  ' * self.level
        print(f'{indent}  fetching {self.url}...')
        a = Article(self.url)
        title = a.title()
        hrefs = list(a.hrefs())
        print(f'{indent}* [{title}] links to {len(hrefs)} articles')
        return title, hrefs

    async def __call__(self, scheduler):
        title, hrefs = await super().__call__(scheduler)
        for href in hrefs:
            if len(scheduler.jobs) < num_articles and href not in scheduler:
                scheduler.add(self.__class__(href, self.level + 1))


events = []
scheduler = Scheduler(workers=num_workers, event_handler=events.append)
scheduler.add(Fetcher(Article.random_url))
asyncio.run(scheduler.run())
if 'plot' in sys.argv:
    from asyncjobs.plot_schedule import plot_schedule

    plot_schedule(
        f'Fetch {num_articles} Wikipedia articles with {num_workers} workers',
        events,
    ).show()



#!/usr/bin/env python3
# Run three simple jobs in sequence
import asyncio
from asyncjobs import Job, Scheduler


# Helper function
def sleep():
    import time

    print(f'{time.ctime()}: Sleep for a second')
    time.sleep(1)
    print(f'{time.ctime()}: Finished sleep')


# Job #1 prints uptime
job1 = Job('#1')
job1.subprocess_argv = ['uptime']

# Job #2 waits for #1 and then sleeps in a thread
job2 = Job('#2', deps={'#1'})
job2.thread_func = sleep

# Job #3 waits for #2 and then prints uptime (again)
job3 = Job('#3', deps={'#2'})
job3.subprocess_argv = ['uptime']

# Run all jobs in the scheduler
s = Scheduler()
for job in [job1, job2, job3]:
    s.add(job)
asyncio.run(s.run())
